from unittest import TestCase
from unittest.mock import Mock, patch

from montecarlodata.common.user import UserService
from montecarlodata.integrations.onboarding.data_lake.presto import PrestoOnboardingService
from montecarlodata.queries.onboarding import TEST_PRESTO_CRED_MUTATION, TEST_S3_CRED_MUTATION
from montecarlodata.utils import GqlWrapper, AwsClientWrapper
from tests.test_base_onboarding import _SAMPLE_BASE_OPTIONS
from tests.test_common_user import _SAMPLE_CONFIG


class PrestoOnBoardingTest(TestCase):
    def setUp(self) -> None:
        self._user_service_mock = Mock(autospec=UserService)
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)

        self._service = PrestoOnboardingService(
            _SAMPLE_CONFIG,
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock
        )

    @patch.object(PrestoOnboardingService, 'onboard')
    def test_presto_sql_flow(self, onboard_mock):
        s3_key = 'testing/test'
        expected_options = {
            'ssl_options': {'mechanism': 'dc-s3', 'cert': 'testing/test', 'skip_verification': False},
            **_SAMPLE_BASE_OPTIONS
        }

        self._service.onboard_presto_sql(**{'cert_s3': s3_key, 'skip_cert_verification': False, **_SAMPLE_BASE_OPTIONS})
        onboard_mock.assert_called_once_with(
            validation_query=TEST_PRESTO_CRED_MUTATION,
            validation_response='testPrestoCredentials',
            connection_type='presto',
            **expected_options
        )

    @patch.object(PrestoOnboardingService, 'onboard')
    def test_presto_s3_flow(self, onboard_mock):
        expected_options = {**_SAMPLE_BASE_OPTIONS, **{'connectionType': 'presto-s3'}}

        self._service.onboard_presto_s3(**_SAMPLE_BASE_OPTIONS)
        onboard_mock.assert_called_once_with(
            validation_query=TEST_S3_CRED_MUTATION,
            validation_response='testS3Credentials',
            connection_type='presto-s3',
            job_types=['query_logs'],
            **expected_options
        )



import copy
from unittest import TestCase
from unittest.mock import Mock, patch, call

import click
from box import Box

from montecarlodata.common.data import AwsGlueAthenaResourceProperties
from montecarlodata.common.resources import CloudResourceService
from montecarlodata.discovery.policy_gen import PolicyDiscoveryService
from tests.test_network_discovery import SAMPLE_DC_PROPS

SAMPLE_BUCKET = 'foo'
SAMPLE_PATH = '/bar/qux'
SAMPLE_GET_TABLES = {
    'TableList': [{'StorageDescriptor': {'Location': f's3://{SAMPLE_BUCKET}{SAMPLE_PATH}'}}], 'NextToken': None}
SAMPLE_GET_WORKGROUP = {
    'Configuration': {'ResultConfiguration': {'OutputLocation': f's3://{SAMPLE_BUCKET}{SAMPLE_PATH}'}}}


class PolicyDiscoveryTest(TestCase):
    def setUp(self) -> None:
        self.maxDiff = None
        self._cloud_resource_service_mock = Mock(autospec=CloudResourceService)

        self._service = PolicyDiscoveryService(cloud_resource_service=self._cloud_resource_service_mock)

    @patch('montecarlodata.discovery.policy_gen.click')
    @patch('montecarlodata.discovery.policy_gen.generate_token')
    @patch('montecarlodata.discovery.policy_gen.read_files')
    @patch('montecarlodata.discovery.policy_gen.render_dumped_json')
    def test_generate_cf_role(self, render_mock, read_mock, generate_mock, click_mock):
        account, external_id, file_vals, template_vals = 'foo', '1234', {'foo': 'bar'}, {'template': {'foo': 'bar'}}
        files, collector_props = ['a', 'b', 'c'], Box({'stackArn': f'arn:partition:service:region:{account}:resource'})

        self._cloud_resource_service_mock.get_and_validate_active_collector.return_value = collector_props
        generate_mock.return_value = external_id
        read_mock.return_value = file_vals
        render_mock.return_value = template_vals

        self._service.generate_cf_role(policy_files=files)

        read_mock.assert_called_once_with(files)
        generate_mock.assert_called_once_with()
        render_mock.assert_called_once_with(
            path='../templates/aws_role.json',
            collector_account_id=account,
            collector_external_id=external_id,
            resource_policies=file_vals
        )
        click_mock.assert_has_calls(
            [
                call.echo('Generating CloudFormation template for creating a role resource.', err=True),
                call.echo(template_vals)
            ]
        )

    @patch('montecarlodata.discovery.policy_gen.complain_and_abort')
    def test_generate_cf_role_without_a_policy(self, complain_mock):
        complain_mock.side_effect = click.exceptions.Abort

        with self.assertRaises(click.exceptions.Abort):
            self._service.generate_cf_role(policy_files=[])
        complain_mock.assert_called_once_with('At least one policy is required.')

    @patch.object(PolicyDiscoveryService, '_get_common_glue_athena_props')
    @patch('montecarlodata.discovery.policy_gen.click')
    @patch('montecarlodata.discovery.policy_gen.render_dumped_json')
    def test_generate_glue_policy(self, render_mock, click_mock, common_mock):
        options, database_name, bucket_names, account_id = {'foo': 'bar'}, {'a', 'b'}, {'c', 'd'}, '1234'
        rendering = {'hello': 'world'}
        props = copy.deepcopy(SAMPLE_DC_PROPS)

        common_mock.return_value = AwsGlueAthenaResourceProperties(
            database_names=database_name,
            bucket_names=bucket_names,
            dc_resource_props=props
        )
        props.resources_client.get_caller_identity.return_value = account_id
        render_mock.return_value = rendering

        self._service.generate_glue_policy(**options)

        common_mock.assert_called_once_with(**options)
        render_mock.assert_called_once_with(
            path='../templates/glue_policy.json',
            region='us-east-1',
            account_id=account_id,
            database_names=database_name,
            data_buckets=bucket_names
        )
        click_mock.assert_has_calls(
            [
                call.echo('Generating policy for Glue.', err=True),
                call.echo(rendering)
            ]
        )

    @patch.object(PolicyDiscoveryService, '_get_workgroup_results')
    @patch.object(PolicyDiscoveryService, '_get_common_glue_athena_props')
    @patch('montecarlodata.discovery.policy_gen.click')
    @patch('montecarlodata.discovery.policy_gen.render_dumped_json')
    def test_generate_athena_policy(self, render_mock, click_mock, common_mock, workgroup_mock):
        workgroup_name = 'workgroup'
        options, bucket_name, path = {'foo': 'bar', 'workgroup_name': workgroup_name}, 'bucket', 'prefix/key'
        database_name, bucket_names, account_id, rendering = {'a', 'b'}, {'c', 'd'}, '1234', {'hello': 'world'}
        props = copy.deepcopy(SAMPLE_DC_PROPS)

        common_mock.return_value = AwsGlueAthenaResourceProperties(database_name, bucket_names, props)
        workgroup_mock.return_value = (bucket_name, path)
        props.resources_client.get_caller_identity.return_value = account_id
        render_mock.return_value = rendering

        self._service.generate_athena_policy(**options)

        common_mock.assert_called_once_with(foo='bar')
        workgroup_mock.assert_called_once_with(workgroup_name=workgroup_name, dc_resource_props=props)
        render_mock.assert_called_once_with(
            path='../templates/athena_policy.json',
            region='us-east-1',
            account_id=account_id,
            database_names=database_name,
            data_buckets=bucket_names,
            result_bucket=bucket_name,
            result_path=path,
            workgroup_name=workgroup_name
        )
        click_mock.assert_has_calls(
            [
                call.echo('Generating policy for Athena.', err=True),
                call.echo(rendering)
            ]
        )

    @patch.object(PolicyDiscoveryService, '_get_buckets')
    def test_get_common_glue_athena_props(self, bucket_mock):
        database_names, bucket_names = ['a', 'b', '*'], ['c', 'd', 'd']
        collector_props, dc_resource_props = Box({'foo': 'bar'}), copy.deepcopy(SAMPLE_DC_PROPS)

        self._cloud_resource_service_mock.get_and_validate_active_collector.return_value = collector_props
        self._cloud_resource_service_mock.get_dc_resource_props.return_value = dc_resource_props

        self.assertEqual(
            self._service._get_common_glue_athena_props(database_names=database_names, bucket_names=bucket_names),
            AwsGlueAthenaResourceProperties(
                database_names={'*'},
                bucket_names={'c', 'd'},
                dc_resource_props=dc_resource_props
            )
        )
        self._cloud_resource_service_mock.get_and_validate_active_collector.assert_called_once_with(dc_id=None)
        self._cloud_resource_service_mock.get_dc_resource_props.assert_called_once_with(
            collector_props=collector_props,
            get_stack_outputs=False,
            get_stack_params=False
        )
        bucket_mock.assert_not_called()

    @patch.object(PolicyDiscoveryService, '_get_buckets')
    def test_get_common_glue_athena_props_with_no_buckets(self, bucket_mock):
        database_names, bucket_names, dc_id = ['a', 'b', '*'], {'c', 'd'}, '1234'
        collector_props, dc_resource_props = Box({'foo': 'bar'}), copy.deepcopy(SAMPLE_DC_PROPS)

        bucket_mock.return_value = bucket_names

        self._cloud_resource_service_mock.get_and_validate_active_collector.return_value = collector_props
        self._cloud_resource_service_mock.get_dc_resource_props.return_value = dc_resource_props

        self.assertEqual(
            self._service._get_common_glue_athena_props(database_names=database_names, dc_id=dc_id),
            AwsGlueAthenaResourceProperties(
                database_names={'*'},
                bucket_names=bucket_names,
                dc_resource_props=dc_resource_props
            )
        )
        self._cloud_resource_service_mock.get_and_validate_active_collector.assert_called_once_with(dc_id=dc_id)
        self._cloud_resource_service_mock.get_dc_resource_props.assert_called_once_with(
            collector_props=collector_props,
            get_stack_outputs=False,
            get_stack_params=False
        )
        bucket_mock.assert_called_once_with(database_names={'*'}, dc_resource_props=dc_resource_props)

    @patch('montecarlodata.discovery.policy_gen.click.progressbar')
    @patch.object(PolicyDiscoveryService, '_get_table_locations')
    def test_get_buckets(self, table_location_mock, click_mock):
        class MockContext:
            def __enter__(self):
                return self

            def __iter__(self):
                for x in database_names:
                    yield x

            def __exit__(self, *args):
                pass

        tables = Box(SAMPLE_GET_TABLES, camel_killer_box=True)
        database_names, locations = {'foo', 'bar'}, {'bucket1', 'bucket2'}
        dc_resource_props = copy.deepcopy(SAMPLE_DC_PROPS)

        table_location_mock.return_value = locations
        click_mock.return_value = MockContext()

        dc_resource_props.resources_client.get_glue_tables.return_value = tables

        self.assertEqual(
            self._service._get_buckets(database_names=database_names, dc_resource_props=dc_resource_props),
            locations
        )
        table_location_mock.assert_has_calls([call(tables=tables.table_list), call(tables=tables.table_list)])
        dc_resource_props.resources_client.get_glue_tables.assert_has_calls(
            [
                call(database_name='bar', page_token=None),
                call(database_name='foo', page_token=None)
            ],
            any_order=True
        )

    def test_handle_wildcards(self):
        self.assertEqual(self._service._handle_wildcards(None), set())
        self.assertEqual(self._service._handle_wildcards([]), set())
        self.assertEqual(self._service._handle_wildcards(['foo', 'foo', 'foo']), {'foo'})
        self.assertEqual(self._service._handle_wildcards(['foo', 'foo', '*']), {'*'})

    @patch('montecarlodata.discovery.policy_gen.click.progressbar')
    def test_get_workgroup_results(self, click_mock):
        workgroup_name, dc_resource_props = 'foo', copy.deepcopy(SAMPLE_DC_PROPS)
        workgroup_details = Box(SAMPLE_GET_WORKGROUP, camel_killer_box=True)

        dc_resource_props.resources_client.get_athena_workgroup.return_value = workgroup_details
        self.assertEqual(
            self._service._get_workgroup_results(workgroup_name, dc_resource_props=dc_resource_props),
            (SAMPLE_BUCKET, SAMPLE_PATH)
        )
        dc_resource_props.resources_client.get_athena_workgroup.assert_called_once_with(workgroup=workgroup_name)

    def test_get_table_locations(self):
        tables = Box(SAMPLE_GET_TABLES, camel_killer_box=True)
        self.assertEqual(list(self._service._get_table_locations(tables=tables.table_list)), [SAMPLE_BUCKET])

    def test_get_table_locations_with_invalid_schemes(self):
        tables = Box(SAMPLE_GET_TABLES, camel_killer_box=True)
        tables.table_list[0].storage_descriptor.location = 'invalid'
        self.assertEqual(list(self._service._get_table_locations(tables=tables.table_list)), [])



from typing import Optional, List, Set, Iterator, io, Tuple
from urllib.parse import urlparse

import click
from box import Box
from click._compat import get_text_stderr

from montecarlodata.common.common import render_dumped_json, read_files, generate_token
from montecarlodata.common.data import DcResourceProperties, AWSArn, AwsGlueAthenaResourceProperties
from montecarlodata.common.resources import CloudResourceService
from montecarlodata.errors import complain_and_abort, manage_errors


class PolicyDiscoveryService:
    ANY_RESOURCE_KEY = '*'

    # Rel template paths. Full path retrieved via pkgutil.
    CF_ROLE_REL_PATH = '../templates/aws_role.json'
    GLUE_POLICY_REL_PATH = '../templates/glue_policy.json'
    ATHENA_POLICY_REL_PATH = '../templates/athena_policy.json'

    def __init__(self, cloud_resource_service: Optional[CloudResourceService] = None, *args, **kwargs):
        self._abort_on_error = True  # Used in decorator
        self._cloud_resource_service = cloud_resource_service or CloudResourceService(*args, **kwargs)

    @manage_errors
    def generate_cf_role(self, policy_files: List[io], dc_id: Optional[str] = None) -> None:
        """
        Generate a CloudFormation template to create a MC compatible IAM role from one or more IAM policy.
        """
        click.echo('Generating CloudFormation template for creating a role resource.', err=True)
        if not len(policy_files) > 0:
            complain_and_abort('At least one policy is required.')

        collector_arn = AWSArn(self._cloud_resource_service.get_and_validate_active_collector(dc_id=dc_id).stackArn)

        click.echo(
            render_dumped_json(
                path=self.CF_ROLE_REL_PATH,
                collector_account_id=collector_arn.account,
                collector_external_id=generate_token(),
                resource_policies=read_files(policy_files)
            )
        )

    @manage_errors
    def generate_glue_policy(self, **kwargs) -> None:
        """
        Generate an IAM compatible policy for Glue from the specified database names or wildcard sequences.

        Bucket names are derived from the database names if not otherwise provided.
        """
        click.echo('Generating policy for Glue.', err=True)
        glue_properties = self._get_common_glue_athena_props(**kwargs)

        click.echo(
            render_dumped_json(
                path=self.GLUE_POLICY_REL_PATH,
                account_id=glue_properties.dc_resource_props.resources_client.get_caller_identity(),
                region=glue_properties.dc_resource_props.resources_region,
                database_names=glue_properties.database_names,
                data_buckets=glue_properties.bucket_names
            )
        )

    @manage_errors
    def generate_athena_policy(self, *, workgroup_name: str, **kwargs) -> None:
        """
        Generate an IAM compatible policy for Athena from the specified database names or wildcard sequences.

        Bucket names are derived from the database names if not otherwise provided. The results bucket is derived from the workgroup.
        """
        click.echo('Generating policy for Athena.', err=True)
        athena_properties = self._get_common_glue_athena_props(**kwargs)

        result_bucket, result_path = self._get_workgroup_results(
            workgroup_name=workgroup_name,
            dc_resource_props=athena_properties.dc_resource_props
        )

        click.echo(
            render_dumped_json(
                path=self.ATHENA_POLICY_REL_PATH,
                account_id=athena_properties.dc_resource_props.resources_client.get_caller_identity(),
                region=athena_properties.dc_resource_props.resources_region,
                database_names=athena_properties.database_names,
                data_buckets=athena_properties.bucket_names,
                result_bucket=result_bucket,
                result_path=result_path,
                workgroup_name=workgroup_name
            )
        )

    def _get_common_glue_athena_props(self, *,
                                      database_names: List[str],
                                      bucket_names: Optional[List[str]] = None,
                                      dc_id: Optional[str] = None, **kwargs) -> AwsGlueAthenaResourceProperties:
        """
        Get resources that are common to Glue and Athena.
        """
        database_names = self._handle_wildcards(database_names)
        bucket_names = self._handle_wildcards(bucket_names)

        dc_resource_props = self._cloud_resource_service.get_dc_resource_props(
            collector_props=self._cloud_resource_service.get_and_validate_active_collector(dc_id=dc_id),
            get_stack_outputs=False,
            get_stack_params=False,
            **kwargs
        )

        if not bucket_names:
            bucket_names = self._get_buckets(database_names=database_names, dc_resource_props=dc_resource_props)

        return AwsGlueAthenaResourceProperties(
            database_names=database_names,
            bucket_names=bucket_names,
            dc_resource_props=dc_resource_props
        )

    def _get_buckets(self, database_names: Set[str], dc_resource_props: DcResourceProperties) -> Set[str]:
        """
        Derive s3 locations (buckets) from provided databases.
        """
        if self.ANY_RESOURCE_KEY in database_names:
            return {self.ANY_RESOURCE_KEY}

        buckets = set()
        with click.progressbar(database_names, label='Looking up buckets details.', file=get_text_stderr()) as dbs:
            for database_name in dbs:
                page_token = None
                while True:
                    tables = dc_resource_props.resources_client.get_glue_tables(
                        database_name=database_name,
                        page_token=page_token
                    )
                    page_token = tables.next_token
                    buckets.update(self._get_table_locations(tables=tables.table_list))

                    if not page_token:
                        break
        return buckets

    def _handle_wildcards(self, sequence: List[str]) -> Set[str]:
        """
        Get unique records and prioritize any wildcards in the sequence.
        """
        sequence = set(sequence) if sequence else set()
        if self.ANY_RESOURCE_KEY in sequence:
            return {self.ANY_RESOURCE_KEY}
        return sequence

    @staticmethod
    def _get_workgroup_results(workgroup_name: str, dc_resource_props: DcResourceProperties) -> Tuple[str, str]:
        """
        Get workgroup result bucket and path.
        """
        with click.progressbar(length=1, label='Looking up workgroup details.', file=get_text_stderr()) as progress:
            workgroup_details = dc_resource_props.resources_client.get_athena_workgroup(workgroup=workgroup_name)

            url = urlparse(workgroup_details.configuration.result_configuration.output_location)
            progress.update(1)
            return url.netloc, url.path if url.path else '/'

    @staticmethod
    def _get_table_locations(tables: Box, valid_scheme_prefix: Optional[str] = 's3') -> Iterator[str]:
        """
        Get locations from s3 for any valid schemes.
        """
        for table in tables:
            url = urlparse(table.storage_descriptor.location)
            if url.scheme.lower().startswith(valid_scheme_prefix):
                yield url.netloc



import copy
import dataclasses
import json
from unittest import TestCase
from unittest.mock import Mock, patch, MagicMock, call, ANY, PropertyMock
from io import StringIO 

import click
from box import Box, BoxList

from montecarlodata.common.data import EventProperties, ResourceProperties, AWSArn, DcResourceProperties, BucketRow
from montecarlodata.common.resources import CloudResourceService
from montecarlodata.common.user import UserService
from montecarlodata.utils import AwsClientWrapper
from tests.helpers import capture_function
from tests.test_common_user import _SAMPLE_CONFIG

_SAMPLE_REGION = 'us-east-1'
_SAMPLE_BUCKET_NAME = 'bucket'
_SAMPLE_BASE_OPTIONS = {'foo': 'bar'}
_SAMPLE_DC_OUTPUT = [{'OutputKey': 'PrivateS3BucketArn', 'OutputValue': f'arn:aws:s3:::{_SAMPLE_BUCKET_NAME}'}]
_SAMPLE_CONNECTION_PATH = 'test'
_SAMPLE_ACCOUNT_ID = '1234'
_SAMPLE_EXTERNAL_ID = '5678'
_SAMPLE_TIME = 'ts'
_SAMPLE_POLICY_FILE = {'hello': 'world'}
_SAMPLE_TRUST_POLICY = {
    'Version': '2012-10-17',
    'Statement': [
        {
            'Effect': 'Allow',
            'Principal': {
                'AWS': f'arn:aws:iam::{_SAMPLE_ACCOUNT_ID}:root'
            },
            'Action': 'sts:AssumeRole',
            'Condition': {
                'StringEquals': {
                    'sts:ExternalId': _SAMPLE_EXTERNAL_ID
                }
            }
        }
    ]
}
_SAMPLE_EMR_NAME = 'name'
_SAMPLE_EMR_ID = 'cluster_id'
_SAMPLE_EMR_ID_2 = 'cluster_id_2'
_SAMPLE_EMR_STATE = 'WAITING'
_SAMPLE_EMR_STATUS = {'State': _SAMPLE_EMR_STATE}
_SAMPLE_EMR_LOG_URI = 's3://bucket/path'
_SAMPLE_EMR_LOG_URI_2 = 's3://bucket_2/path'

_SAMPLE_BUCKETS_FILE = """bucket_name,prefix,suffix
foo,,
bar,prefix1,
baz,,suffix1
qux,prefix2,suffix2
"""

_SAMPLE_EVENT_PAYLOAD = {
    'bucket_name': 'mcd-test-events',
    'collector_aws_profile': 'dev',
    'prefix': None,
    'suffix': None,
    'topic_arn': None,
    'bucket_aws_profile': None,
    'event_type': 'metadata'
}
_SAMPLE_BUCKETS_FILE_EVENT_PAYLOAD = {
    'buckets_filename': 'buckets.csv',
    'collector_aws_profile': 'dev',
    'prefix': None,
    'suffix': None,
    'topic_arn': None,
    'bucket_aws_profile': None,
    'event_type': 'query-logs'
}
_SAMPLE_COLLECTOR = {
    'uuid': 'e34eaf86-092d-47ac-97c3-cfa8b6a6e7fe',
    'stackArn': 'arn:aws:cloudformation:us-east-1:123456789:stack/loki/99',
    'active': True,
    'customerAwsAccountId': '123456789',
    'templateVersion': '42',
    'codeVersion': '2042',
    'lastUpdated': None
}

_SAMPLE_DC_CLIENT = Mock()
_SAMPLE_CUST_RESOURCES_CLIENT = Mock()

_SAMPLE_RESOURCE_PROPERTY_1 = ResourceProperties(
    bucket_name='mcd-test-events',
    client=copy.deepcopy(_SAMPLE_CUST_RESOURCES_CLIENT),
    account_id='123456789',
    region='us-east-1'
)
_SAMPLE_RESOURCE_PROPERTY_2 = ResourceProperties(
    bucket_name='mcd-test-events-us-east-2',
    client=copy.deepcopy(_SAMPLE_CUST_RESOURCES_CLIENT),
    account_id='123456789',
    region='us-east-2'
)

_SAMPLE_RESOURCE_PROPERTIES = [
    _SAMPLE_RESOURCE_PROPERTY_1,
    _SAMPLE_RESOURCE_PROPERTY_2
]
_SAMPLE_EVENT_DETAILS = EventProperties(
    event_type='metadata',
    event_queue_arn='arn:aws:sqs:us-east-1:123456789:loki-MetadataEventQueue-42',
    collection_region='us-east-1',
    collection_account_id='123456789',
    collection_client=_SAMPLE_DC_CLIENT,
    resource_properties=_SAMPLE_RESOURCE_PROPERTIES
)


class CloudResourcesTest(TestCase):
    def setUp(self) -> None:
        self._user_service_mock = Mock(autospec=UserService)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)

        self._service = CloudResourceService(
            _SAMPLE_CONFIG,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock
        )

    def tearDown(self) -> None:
        _SAMPLE_DC_CLIENT.reset_mock()
        _SAMPLE_CUST_RESOURCES_CLIENT.reset_mock()

    @patch.object(CloudResourceService, '_read_json')
    def test_create_role_with_invalid_policy(self, mock_read):
        mock_read.side_effect = json.decoder.JSONDecodeError
        self._user_service_mock.active_collector = {'customerAwsAccountId': _SAMPLE_ACCOUNT_ID}

        with self.assertRaises(click.exceptions.Abort):
            self._service.create_role(path_to_policy_doc='invalid')

    @patch.object(CloudResourceService, '_read_json')
    def test_create_role_with_missing_values(self, mock_read):
        mock_read.return_value = _SAMPLE_POLICY_FILE
        self._user_service_mock.active_collector = {}
        with self.assertRaises(click.exceptions.Abort):
            self._service.create_role(path_to_policy_doc='foo')

    @patch.object(CloudResourceService, '_read_json')
    @patch.object(CloudResourceService, '_generate_random_token')
    @patch('time.time', MagicMock(return_value=_SAMPLE_TIME))
    def test_create_role(self, mock_token, mock_read):
        path_to_policy_doc = 'foo'
        mock_read.return_value = _SAMPLE_POLICY_FILE
        mock_token.return_value = _SAMPLE_EXTERNAL_ID
        self._user_service_mock.active_collector = {'customerAwsAccountId': _SAMPLE_ACCOUNT_ID}

        self._service.create_role(path_to_policy_doc=path_to_policy_doc)

        mock_read.assert_called_once_with(path_to_policy_doc)
        mock_token.assert_called_once_with()

        self._aws_wrapper_mock.create_role.assert_called_once_with(
            role_name=f'monte-carlo-integration-role-{_SAMPLE_TIME}',
            trust_policy=json.dumps(_SAMPLE_TRUST_POLICY),
            tags=self._service._MCD_TAGS
        )

        self._aws_wrapper_mock.attach_inline_policy.assert_called_once_with(
            role_name=f'monte-carlo-integration-role-{_SAMPLE_TIME}',
            policy_name=f'monte-carlo-integration-cli-policy-{_SAMPLE_TIME}',
            policy_doc=json.dumps(_SAMPLE_POLICY_FILE)
        )

    def test_generate_policy(self):
        policy = self._service._generate_trust_policy(account_id=_SAMPLE_ACCOUNT_ID, external_id=_SAMPLE_EXTERNAL_ID)

        self.assertEqual(policy, json.dumps(_SAMPLE_TRUST_POLICY))

    def test_list_emr_clusters(self):
        self._aws_wrapper_mock.get_emr_cluster_details.return_value = [{
            'Name': _SAMPLE_EMR_NAME,
            'Id': _SAMPLE_EMR_ID,
            'Status': _SAMPLE_EMR_STATUS,
            'LogUri': _SAMPLE_EMR_LOG_URI
        }]

        std_out = capture_function(self._service.list_emr_clusters, {'table_format': 'plain'}).std_out

        expected_output = ('Name    Id          State    LogUri\n'
                           f'{_SAMPLE_EMR_NAME}    {_SAMPLE_EMR_ID}  {_SAMPLE_EMR_STATE}  {_SAMPLE_EMR_LOG_URI}')
        self.assertEqual(std_out.getvalue().strip(), expected_output)

    def test_list_emr_clusters_no_wait(self):
        self._aws_wrapper_mock.get_emr_cluster_details.return_value = [{
            'Name': _SAMPLE_EMR_NAME,
            'Id': _SAMPLE_EMR_ID,
            'Status': _SAMPLE_EMR_STATUS,
            'LogUri': _SAMPLE_EMR_LOG_URI
        }]

        std_out = capture_function(self._service.list_emr_clusters, {'no_grid': True}).std_out
        expected_output = '  '.join([_SAMPLE_EMR_NAME, _SAMPLE_EMR_ID, _SAMPLE_EMR_STATE, _SAMPLE_EMR_LOG_URI])
        self.assertEqual(std_out.getvalue().strip().strip(), expected_output)

    def test_list_emr_clusters_only_log_locations(self):
        self._aws_wrapper_mock.get_emr_cluster_details.return_value = [
            {
                'Name': _SAMPLE_EMR_NAME,
                'Id': _SAMPLE_EMR_ID,
                'Status': _SAMPLE_EMR_STATUS,
                'LogUri': _SAMPLE_EMR_LOG_URI
            },
            {
                'Name': _SAMPLE_EMR_NAME,
                'Id': _SAMPLE_EMR_ID_2,
                'Status': _SAMPLE_EMR_STATUS,
                'LogUri': _SAMPLE_EMR_LOG_URI_2
            }
        ]

        std_out = capture_function(self._service.list_emr_clusters, {'only_log_locations': True}).std_out
        self.assertEqual(set(std_out.getvalue().strip().split('\n')), {_SAMPLE_EMR_LOG_URI, _SAMPLE_EMR_LOG_URI_2})

    def test_list_emr_clusters_no_clusters(self):
        self._aws_wrapper_mock.get_emr_cluster_details.return_value = []
        self._user_service_mock.active_collector = {'customerAwsAccountId': _SAMPLE_ACCOUNT_ID}

        std_out = capture_function(self._service.list_emr_clusters, {'table_format': 'plain'}).std_out
        self.assertEqual(std_out.getvalue().strip(), 'Name    Id    State    LogUri')

    @patch('montecarlodata.common.resources.click')
    @patch.object(CloudResourceService, '_initialize_events')
    @patch.object(CloudResourceService, '_create_or_update_event_topic')
    @patch.object(CloudResourceService, '_update_event_q')
    @patch.object(CloudResourceService, '_create_topic_subscription')
    @patch.object(CloudResourceService, '_create_event_notification')
    def test_add_events_flow(self, mock_notification, mock_topic_sub, mock_q, mock_topic, mock_setup, mock_click):
        self._user_service_mock.get_collector.return_value = Box(_SAMPLE_COLLECTOR)
        mock_setup.return_value = _SAMPLE_EVENT_DETAILS

        self._service.add_events(**_SAMPLE_EVENT_PAYLOAD)

        self._user_service_mock.get_collector.assert_called_once_with(dc_id=None)

        mock_setup.assert_called_once_with(collector_props=_SAMPLE_COLLECTOR, **_SAMPLE_EVENT_PAYLOAD)
        mock_topic.assert_called_once_with(event_details=_SAMPLE_EVENT_DETAILS, skip_prompts=False)
        mock_q.assert_called_once_with(event_details=_SAMPLE_EVENT_DETAILS, skip_prompts=False)
        mock_topic_sub.assert_called_once_with(event_details=_SAMPLE_EVENT_DETAILS, skip_prompts=False)
        mock_notification.assert_called_once_with(event_details=_SAMPLE_EVENT_DETAILS, skip_prompts=False)

        mock_click.assert_has_calls(
            [
                call.echo('[Gathering collector and resource details]'),
                call.echo('\nProcess complete. Have a nice day!')
            ]
        )

    def test_add_events_with_an_inactive_collector(self):
        dc_id = 'thor'
        inactive_collector = Box(copy.deepcopy(_SAMPLE_COLLECTOR))
        inactive_collector.active = False

        self._user_service_mock.get_collector.return_value = inactive_collector
        with self.assertRaises(click.exceptions.Abort):
            self._service.add_events(**_SAMPLE_EVENT_PAYLOAD, **{'dc_id': dc_id})
        self._user_service_mock.get_collector.assert_called_once_with(dc_id=dc_id)

    @patch('montecarlodata.common.resources.AwsClientWrapper')
    @patch.object(CloudResourceService, '_get_relevant_q')
    def test_initialize_events(self, mock_get_q, aws_wrapper):
        derived_bucket_location, dc_outputs, q_arn = 'thor', {'foo': 'bar'}, 'arn:aws:sqs:us-east-1:12345:odin'
        aws_wrapper().get_bucket_location.return_value = derived_bucket_location
        aws_wrapper().region = derived_bucket_location
        aws_wrapper().get_stack_outputs.return_value = dc_outputs
        aws_wrapper().get_caller_identity.return_value = AWSArn(q_arn).account

        mock_get_q.return_value = q_arn

        answer = self._service._initialize_events(collector_props=Box(_SAMPLE_COLLECTOR), **_SAMPLE_EVENT_PAYLOAD)
        expected = EventProperties(
                event_type=_SAMPLE_EVENT_PAYLOAD['event_type'],
                event_queue_arn=q_arn,
                collection_region=AWSArn(q_arn).region,
                collection_account_id=AWSArn(_SAMPLE_COLLECTOR['stackArn']).account,
                collection_client=ANY,
                resource_properties=[
                    ResourceProperties(
                        account_id=AWSArn(q_arn).account,
                        region=derived_bucket_location,
                        bucket_name=_SAMPLE_EVENT_PAYLOAD['bucket_name'],
                        client=ANY
                    )
                ]
            )
        self.assertEqual(answer, expected)

        aws_wrapper.assert_has_calls(
            [
                call(profile_name=_SAMPLE_EVENT_PAYLOAD['collector_aws_profile'],
                     region_name=_SAMPLE_CONFIG.aws_region),
                call().get_stack_outputs(stack_id=_SAMPLE_COLLECTOR['stackArn']),
                call(profile_name=_SAMPLE_CONFIG.aws_profile),
                call().get_bucket_location(bucket_name='mcd-test-events'),
                call(profile_name=_SAMPLE_CONFIG.aws_profile, region_name=derived_bucket_location),
                call().get_caller_identity()
            ]
        )
        mock_get_q.assert_called_once_with('metadata', dc_outputs)

    @patch('montecarlodata.common.resources.AwsClientWrapper')
    @patch.object(CloudResourceService, '_get_relevant_q')
    @patch.object(CloudResourceService, 'parse_buckets_file')
    def test_initialize_events_from_buckets_file(self, mock_parse_buckets, mock_get_q, aws_wrapper):
        derived_bucket_location, dc_outputs, q_arn = 'thor', {'foo': 'bar'}, 'arn:aws:sqs:us-east-1:12345:odin'
        aws_wrapper().get_bucket_location.return_value = derived_bucket_location
        aws_wrapper().region = derived_bucket_location
        aws_wrapper().get_stack_outputs.return_value = dc_outputs
        aws_wrapper().get_caller_identity.return_value = AWSArn(q_arn).account

        mock_get_q.return_value = q_arn
        mock_parse_buckets.return_value = [
            BucketRow(bucket_name="foo"),
            BucketRow(bucket_name="bar"),
            BucketRow(bucket_name="baz")
        ]

        answer = self._service._initialize_events(collector_props=Box(_SAMPLE_COLLECTOR), **_SAMPLE_BUCKETS_FILE_EVENT_PAYLOAD)
        expected = EventProperties(
                event_type=_SAMPLE_BUCKETS_FILE_EVENT_PAYLOAD['event_type'],
                event_queue_arn=q_arn,
                collection_region=AWSArn(q_arn).region,
                collection_account_id=AWSArn(_SAMPLE_COLLECTOR['stackArn']).account,
                collection_client=ANY,
                resource_properties=[
                    ResourceProperties(
                        account_id=AWSArn(q_arn).account,
                        region=derived_bucket_location,
                        bucket_name='foo',
                        client=ANY
                    ),
                    ResourceProperties(
                        account_id=AWSArn(q_arn).account,
                        region=derived_bucket_location,
                        bucket_name='bar',
                        client=ANY
                    ),
                    ResourceProperties(
                        account_id=AWSArn(q_arn).account,
                        region=derived_bucket_location,
                        bucket_name='baz',
                        client=ANY
                    )
                ]
            )

        self.assertEqual(answer, expected)

        aws_wrapper.assert_has_calls(
            [
                call(profile_name=_SAMPLE_EVENT_PAYLOAD['collector_aws_profile'],
                     region_name=_SAMPLE_CONFIG.aws_region),
                call().get_stack_outputs(stack_id=_SAMPLE_COLLECTOR['stackArn']),
                call(profile_name=_SAMPLE_CONFIG.aws_profile),
                call().get_bucket_location(bucket_name='foo'),
                call(profile_name=_SAMPLE_CONFIG.aws_profile, region_name=derived_bucket_location),
                call().get_bucket_location(bucket_name='bar'),
                call().get_bucket_location(bucket_name='baz'),
                call().get_caller_identity(),
                call().get_caller_identity(),
                call().get_caller_identity()
            ]
        )
        mock_get_q.assert_called_once_with('query-logs', dc_outputs)

    @patch('builtins.open')
    def test_parse_buckets_file(self, mock_open):
        mock_open.return_value = StringIO(_SAMPLE_BUCKETS_FILE)
        answer = self._service.parse_buckets_file("test_filename")
        expected = [
            BucketRow(bucket_name='foo', prefix='', suffix=''),
            BucketRow(bucket_name='bar', prefix='prefix1', suffix=''),
            BucketRow(bucket_name='baz', prefix='', suffix='suffix1'),
            BucketRow(bucket_name='qux', prefix='prefix2', suffix='suffix2')
        ]

        self.assertEqual(answer, expected)

    def test_get_relevant_q(self):
        m_out, q_out = 'loki', 'thor'
        sample_outputs = [
            {'OutputKey': 'MetadataEventQueue', 'OutputValue': m_out, 'Description': 'Metadata event queue ARN'},
            {'OutputKey': 'PublicIP', 'OutputValue': '127.0.0.1', 'Description': 'Does stuff'},
            {'OutputKey': 'QueryLogEventQueue', 'OutputValue': q_out, 'Description': 'Query log event queue ARN'}
        ]

        self.assertEqual(self._service._get_relevant_q(event_type='metadata', outputs=BoxList(sample_outputs)), m_out)
        self.assertEqual(self._service._get_relevant_q(event_type='query-logs', outputs=BoxList(sample_outputs)), q_out)
        with self.assertRaises(click.exceptions.Abort):
            self._service._get_relevant_q(event_type='invalid', outputs=BoxList(sample_outputs))  # bad event type
        with self.assertRaises(click.exceptions.Abort):
            self._service._get_relevant_q(event_type='metadata', outputs=Box({}))  # missing event config

    @patch('montecarlodata.common.resources.prompt_connection')
    @patch.object(CloudResourceService, '_extend_topic_policy')
    @patch.object(CloudResourceService, '_is_preexisting_topic_policy')
    @patch.object(CloudResourceService, '_get_preexisting_topic_policy')
    @patch.object(CloudResourceService, '_save_event_properties')
    def test_create_or_update_event_topic(self, mock_save_event_properties, mock_get_preexisting, mock_is_preexisting, mock_extend_policy, mock_prompt):
        attributes = Box({'Condition': {'StringEquals': {'aws:SourceArn': 'foo'}}})

        events = copy.deepcopy(_SAMPLE_EVENT_DETAILS)
        for resource in events.resource_properties:
            region = resource.region
            topic_name = f'monte-carlo-data-{region}-metadata-events-topic'
            topic_arn = f'arn:aws:sns:{region}:12345:monte-carlo-data-{region}-metadata-events-topic'
            resource.client.create_sns_topic.return_value = topic_arn
            resource.client.get_topic_attributes.return_value = attributes
        
        mock_get_preexisting.return_value = attributes
        mock_is_preexisting.return_value = False
        mock_extend_policy.return_value = attributes

        self._service._create_or_update_event_topic(event_details=events, skip_prompts=True)

        mock_is_preexisting.assert_has_calls(
            [
                call(
                    topic_policy_id='monte-carlo-data-metadata-loki-MetadataEventQueue-42-ID',
                    topic_attributes=attributes
                ),
                call( 
                    topic_policy_id='monte-carlo-data-metadata-loki-MetadataEventQueue-42-ID',
                    topic_attributes=attributes
                )
            ]
        )
        mock_extend_policy.assert_has_calls(
            [
                call(
                    topic_attributes=Box({"Condition": {"StringEquals": {"aws:SourceArn": "foo"}}}),
                    new_statements=[
                        {
                            "Sid": "monte-carlo-data-metadata-loki-MetadataEventQueue-42-ID",
                            "Effect": "Allow",
                            "Principal": {"AWS": "123456789"},
                            "Action": "sns:Subscribe",
                            "Resource": "arn:aws:sns:us-east-1:12345:monte-carlo-data-us-east-1-metadata-events-topic",
                        }
                    ],
                    update_statements={
                        "monte-carlo-data-metadata-monte-carlo-data-us-east-1-metadata-events-topic-ID": Box(
                            {
                                "Condition": {
                                    "StringEquals": {
                                        "aws:SourceArn": ["foo", "arn:aws:s3:::mcd-test-events"]
                                    }
                                }
                            }
                        )
                    },
                ),
                call(
                    topic_attributes=Box({"Condition": {"StringEquals": {"aws:SourceArn": "foo"}}}),
                    new_statements=[
                        {
                            "Sid": "monte-carlo-data-metadata-loki-MetadataEventQueue-42-ID",
                            "Effect": "Allow",
                            "Principal": {"AWS": "123456789"},
                            "Action": "sns:Subscribe",
                            "Resource": "arn:aws:sns:us-east-2:12345:monte-carlo-data-us-east-2-metadata-events-topic",
                        }
                    ],
                    update_statements={
                        "monte-carlo-data-metadata-monte-carlo-data-us-east-2-metadata-events-topic-ID": Box(
                            {
                                "Condition": {
                                    "StringEquals": {
                                        "aws:SourceArn": [
                                            "foo",
                                            "arn:aws:s3:::mcd-test-events-us-east-2",
                                        ]
                                    }
                                }
                            }
                        )
                    },
                )            
            ]
        )

        for resource in events.resource_properties:
            resource.client.set_topic_attributes.assert_called_once_with(
                arn=resource.topic_arn,
                name='Policy',
                value=attributes
            )
            resource.client.get_topic_attributes.assert_called_once_with(arn=resource.topic_arn)
            resource.client.create_sns_topic.assert_called_once_with(name=resource.topic_name)
        mock_prompt.assert_has_calls(
            [
                call(
                    message=f"Create topic 'monte-carlo-data-us-east-1-metadata-events-topic' in 'us-east-1' (123456789)? "
                            "This is an idempotent operation.",
                    skip_prompt=True
                ),
                call(message='Please confirm', skip_prompt=True),
                call(
                    message="Create topic 'monte-carlo-data-us-east-2-metadata-events-topic' in 'us-east-2' (123456789)? "
                            "This is an idempotent operation.",
                    skip_prompt=True,
                ),                
                call(message='Please confirm', skip_prompt=True)
            ]
        )

        mock_save_event_properties.assert_called()

    def test_create_or_update_event_topic_arn_invalid(self):
        events = copy.deepcopy(_SAMPLE_EVENT_DETAILS)

        with self.assertRaises(click.exceptions.Abort):
            for resource in events.resource_properties:
                resource.topic_arn = 'foo'
            self._service._create_or_update_event_topic(event_details=events, skip_prompts=True)  # invalid ARN

    def test_create_or_update_event_topic_region_invalid(self):
        events = copy.deepcopy(_SAMPLE_EVENT_DETAILS)

        with self.assertRaises(click.exceptions.Abort):
            for resource in events.resource_properties:
                resource.topic_arn = 'arn:aws:sns:us-west-1:12345:monte-carlo-data-us-east-1-metadata-events-topic'
            self._service._create_or_update_event_topic(event_details=events, skip_prompts=True)  # region mismatch

    def test_preexisting_topic_policy(self):
        sid, topic_attributes = 'foo', Box({'Attributes': {'Policy': '{"Statement": []}'}})
        self.assertFalse(self._service._is_preexisting_topic_policy(sid, topic_attributes))

        topic_attributes = Box({'Attributes': {'Policy': '{"Statement": [{"Sid": "foo"}]}'}})
        self.assertTrue(self._service._is_preexisting_topic_policy(sid, topic_attributes))

    def test_extend_topic_policy(self):
        topic_attributes = Box({'Attributes': {'Policy': '{"Statement": [{"Sid": "foo"}]}'}})
        new_statements = [{'x': 'y'}]
        self.assertEqual(
            self._service._extend_topic_policy(topic_attributes, new_statements, update_statements=[]),
            '{"Statement": [{"Sid": "foo"}, {"x": "y"}]}'
        )

    def test_extend_topic_policy_with_updates(self):
        topic_attributes = Box({'Attributes': {'Policy': '{"Statement": [{"Sid": "foo"}]}'}})
        new_statements = [{'x': 'y'}]
        update_statements = {'foo': {'Sid': 'foo', 'bar': 'baz'}}
        self.assertEqual(
            self._service._extend_topic_policy(topic_attributes, new_statements, update_statements=update_statements),
            '{"Statement": [{"Sid": "foo", "bar": "baz"}, {"x": "y"}]}'
        )

    @patch('montecarlodata.common.resources.prompt_connection')
    @patch.object(CloudResourceService, '_save_event_properties')
    def test_update_event_q(self, mock_save_event_properties, mock_prompt):
        name = 'loki-MetadataEventQueue-42'
        _SAMPLE_DC_CLIENT.get_q_attributes.return_value = None

        events = copy.deepcopy(_SAMPLE_EVENT_DETAILS)
        for resource in events.resource_properties:
            resource.topic_arn = f'arn:aws:sns:{resource.region}:12345:monte-carlo-data-{resource.region}-metadata-events-topic'

        self._service._update_event_q(event_details=events, skip_prompts=True)

        events.collection_client.set_q_attributes.assert_called_once_with(
            name=name,
            attributes={
                'Policy': '{"Version": "2008-10-17", "Statement": [{"Sid": "__owner", "Effect": "Allow", "Principal": '
                          '{"AWS": "arn:aws:iam::123456789:root"}, "Action": "SQS:*", "Resource": '
                          '"arn:aws:sqs:us-east-1:123456789:loki-MetadataEventQueue-42"}, '
                          '{"Sid": "__sender", "Effect": "Allow", "Principal": {"AWS": "*"}, "Action": '
                          '"SQS:SendMessage", "Resource": "arn:aws:sqs:us-east-1:123456789:loki-MetadataEventQueue-42",'
                          ' "Condition": {"ArnLike": {"aws:SourceArn": '
                          '["arn:aws:sns:us-east-1:12345:monte-carlo-data-us-east-1-metadata-events-topic",'
                          ' "arn:aws:sns:us-east-2:12345:monte-carlo-data-us-east-2-metadata-events-topic"]}}}]}'
            }
        )
        events.collection_client.get_q_attributes.assert_called_once_with(name=name, attributes=['Policy'])
        mock_prompt.assert_called_once_with(message='Please confirm', skip_prompt=True)
        mock_save_event_properties.assert_called()

    @patch('montecarlodata.common.resources.prompt_connection')
    @patch.object(CloudResourceService, '_extend_q_policy')
    @patch.object(CloudResourceService, '_save_event_properties')
    def test_update_event_q_with_existing(self, mock_save_event_properties, mock_extend, mock_prompt):
        name, mock_extend.return_value = 'loki-MetadataEventQueue-42', {'foo': 'bar'}

        _SAMPLE_DC_CLIENT.get_q_attributes.return_value = Box({'Attributes': {"Policy": '{}'}})
        self._service._update_event_q(event_details=_SAMPLE_EVENT_DETAILS, skip_prompts=True)

        _SAMPLE_DC_CLIENT.get_q_attributes.assert_called_once_with(name=name, attributes=['Policy'])
        _SAMPLE_DC_CLIENT.set_q_attributes.assert_called_once_with(name=name, attributes={'Policy': '{"foo": "bar"}'})
        mock_prompt.assert_called_once_with(message='Please confirm', skip_prompt=True)
        mock_save_event_properties.assert_called()

    @patch('montecarlodata.common.resources.prompt_connection')
    @patch.object(CloudResourceService, '_extend_q_policy')
    def test_update_event_q_with_no_delta(self, mock_extend, mock_prompt):
        name, mock_extend.return_value = 'loki-MetadataEventQueue-42', {'foo': 'bar'}

        _SAMPLE_DC_CLIENT.get_q_attributes.return_value = Box({'Attributes': {"Policy": '{"foo": "bar"}'}})
        self._service._update_event_q(event_details=_SAMPLE_EVENT_DETAILS, skip_prompts=True)

        _SAMPLE_DC_CLIENT.get_q_attributes.assert_called_once_with(name=name, attributes=['Policy'])
        _SAMPLE_DC_CLIENT.set_q_attributes.assert_not_called()
        mock_prompt.assert_not_called()

    def test_extend_q_policy(self):
        events = copy.deepcopy(_SAMPLE_EVENT_DETAILS)
        for resource in events.resource_properties:
            resource.topic_arn = 'bar'

        # Existing is string
        existing_policy = Box({'Statement': [{'Sid': '__sender', 'Condition': {'ArnLike': {'aws:SourceArn': 'foo'}}}]})
        self.assertEqual(
            self._service._extend_q_policy(existing_policy=existing_policy, event_details=events, topic_arns=['bar']),
            {'Statement': [{'Sid': '__sender', 'Condition': {'ArnLike': {'aws:SourceArn': ['foo', 'bar']}}}]}
        )

        # Existing is list
        existing_policy = Box({'Statement': [{'Sid': '__sender', 'Condition': {'ArnLike': {'aws:SourceArn': ['foo', 'qux']}}}]})
        self.assertEqual(
            self._service._extend_q_policy(existing_policy=existing_policy, event_details=events, topic_arns=['bar']),
            {'Statement': [{'Sid': '__sender', 'Condition': {'ArnLike': {'aws:SourceArn': ['foo', 'qux', 'bar']}}}]}
        )

        # Existing has duplicate
        existing_policy = Box({'Statement': [{'Sid': '__sender', 'Condition': {'ArnLike': {'aws:SourceArn': 'bar'}}}]})
        self.assertEqual(
            self._service._extend_q_policy(existing_policy=existing_policy, event_details=events, topic_arns=['bar']),
            {'Statement': [{'Sid': '__sender', 'Condition': {'ArnLike': {'aws:SourceArn': 'bar'}}}]}
        )
        existing_policy = Box({'Statement': [{'Sid': '__sender', 'Condition': {'ArnLike': {'aws:SourceArn': ['foo', 'bar']}}}]})
        self.assertEqual(
            self._service._extend_q_policy(existing_policy=existing_policy, event_details=events, topic_arns=['foo', 'bar']),
            {'Statement': [{'Sid': '__sender', 'Condition': {'ArnLike': {'aws:SourceArn': ['foo', 'bar']}}}]}
        )

        # Incompatible format
        with self.assertRaises(click.exceptions.Abort):
            self._service._extend_q_policy(existing_policy=Box({'Statement': [{'Sid': 'unk'}]}), event_details=events, topic_arns=['bar'])

    @patch('montecarlodata.common.resources.prompt_connection')
    @patch.object(CloudResourceService, '_merge_event_notifications')
    @patch.object(CloudResourceService, '_save_event_properties')
    def test_create_event_notification(self, mock_save_event_properties, mock_merge_events, mock_prompt):
        events = copy.deepcopy(_SAMPLE_EVENT_DETAILS)
        new_config, existing_config = {'foo': 'bar'}, {'bar': 'foo'}

        for resource in events.resource_properties:
            resource.topic_arn = 'arn:aws:sns:us-east-1:12345:monte-carlo-data-us-east-1-metadata-events-topic'
            resource.prefix = 'lake'
            resource.client.get_bucket_event_config.return_value = existing_config

        mock_merge_events.return_value = new_config
        self._service._create_event_notification(event_details=events, skip_prompts=True)


        expected_merge_events_calls = [
            call(            
                setting_id='monte-carlo-data-us-east-1-metadata-events-topic-lake-notification',
                existing_config=existing_config,
                resource_details=resource
            ) for resource in events.resource_properties
        ]
        self.assertEqual(mock_merge_events.call_args_list, expected_merge_events_calls)

        for resource in events.resource_properties:
           resource.client.get_bucket_event_config.assert_called_once_with(name=resource.bucket_name)
           resource.client.set_bucket_event_config.assert_called_once_with(name=resource.bucket_name, notification_config=new_config)

        expected_prompt_calls = [
            call(            
                message='Please confirm', skip_prompt=True
            ) for _ in events.resource_properties
        ]
        self.assertEqual(mock_prompt.call_args_list, expected_prompt_calls)

        mock_save_event_properties.assert_called()

    @patch('montecarlodata.common.resources.prompt_connection')
    @patch.object(CloudResourceService, '_merge_event_notifications')
    @patch.object(CloudResourceService, '_save_event_properties')
    def test_create_event_notification_with_no_delta(self, mock_save_event_properties, mock_merge_events, mock_prompt):
        events = copy.deepcopy(_SAMPLE_EVENT_DETAILS)
        new_config, existing_config = {'foo': 'bar'}, {'foo': 'bar'}

        for resource in events.resource_properties:
            resource.topic_arn = 'arn:aws:sns:us-east-1:12345:monte-carlo-data-us-east-1-metadata-events-topic'
            resource.client.get_bucket_event_config.return_value = existing_config

        mock_merge_events.return_value = new_config
        self._service._create_event_notification(event_details=events, skip_prompts=True)
        expected_merge_events_calls = [
            call(            
                setting_id='monte-carlo-data-us-east-1-metadata-events-topic-notification',
                existing_config=existing_config,
                resource_details=resource
            ) for resource in events.resource_properties
        ]
        self.assertEqual(mock_merge_events.call_args_list, expected_merge_events_calls)

        for resource in events.resource_properties:
            resource.client.get_bucket_event_config.assert_called_once_with(name=resource.bucket_name)
            resource.client.set_bucket_event_config.assert_not_called()

        mock_prompt.assert_not_called()
        mock_save_event_properties.assert_called()

    def test_merge_event_notifications(self):
        setting_id = 'foo'
        events = copy.deepcopy(_SAMPLE_EVENT_DETAILS)


        for resource in events.resource_properties:
            resource.topic_arn = f'arn:aws:sns:{resource.region}:12345:monte-carlo-data-{resource.region}-metadata-events-topic'
            expected_config = {
                'Id': setting_id,
                'TopicArn': resource.topic_arn,
                'Events': ['s3:ObjectCreated:*', 's3:ObjectRemoved:*'],
                'Filter': {'Key': {'FilterRules': [{'Name': 'Prefix', 'Value': ''}, {'Name': 'Suffix', 'Value': ''}]}}
            }
            # ID already exists
            existing_setting = {'TopicConfigurations': [{'Id': setting_id}]}
            self.assertEqual(
                self._service._merge_event_notifications(setting_id, existing_setting, resource),
                existing_setting
            )

            # Topic already exits
            existing_setting = {'TopicConfigurations': [{'Id': 'bar'}], 'QueueConfigurations': [{'foo': 'bar'}]}
            self.assertEqual(
                self._service._merge_event_notifications(setting_id, existing_setting, resource),
                {'TopicConfigurations': [{'Id': 'bar'}, expected_config], 'QueueConfigurations': [{'foo': 'bar'}]}
            )

            # No topic config
            self.assertEqual(
                self._service._merge_event_notifications(setting_id, {}, resource),
                {'TopicConfigurations': [expected_config]}
            )

            existing_setting = {'QueueConfigurations': [{'foo': 'bar'}]}
            self.assertEqual(
                self._service._merge_event_notifications(setting_id, existing_setting, resource),
                {'TopicConfigurations': [expected_config], 'QueueConfigurations': [{'foo': 'bar'}]}
            )

    @patch('montecarlodata.common.resources.prompt_connection')
    @patch.object(CloudResourceService, '_save_event_properties')
    def test_create_topic_subscription(self, mock_save_event_properties, mock_prompt):
        events = copy.deepcopy(_SAMPLE_EVENT_DETAILS)
        for resource in events.resource_properties:
            resource.client.list_topic_subscriptions.return_value = None
        self._service._create_topic_subscription(event_details=events, skip_prompts=True)

        for resource in events.resource_properties:
            resource.client.list_topic_subscriptions.assert_called_once_with(
                arn=resource.topic_arn
            )
            resource.client.subscribe_to_topic.assert_called_once_with(
                arn=resource.topic_arn,
                endpoint=events.event_queue_arn,
                attributes={'RawMessageDelivery': 'true'}
            )

        mock_prompt.assert_has_calls([
            call(message='Please confirm', skip_prompt=True),
            call(message='Please confirm', skip_prompt=True),
        ])
        mock_save_event_properties.assert_called()

    @patch('montecarlodata.common.resources.prompt_connection')
    @patch.object(CloudResourceService, '_save_event_properties')
    def test_create_topic_subscription_with_preexisting_endpoint(self, mock_save_event_properties, mock_prompt):
        events = copy.deepcopy(_SAMPLE_EVENT_DETAILS)
        for resource in events.resource_properties:
            resource.client.list_topic_subscriptions.return_value = [
                [
                    {
                        'SubscriptionArn': 'odin',
                        'Owner': '12345',
                        'Protocol': 'sqs',
                        'Endpoint': events.event_queue_arn,
                        'TopicArn': resource.topic_arn
                    }
                ]
            ]

        self._service._create_topic_subscription(event_details=events, skip_prompts=True)

        for resource in events.resource_properties:
            resource.client.list_topic_subscriptions.assert_called_once_with(
                arn=resource.topic_arn
            )
            _SAMPLE_CUST_RESOURCES_CLIENT.subscribe_to_topic.cassert_not_called()

        mock_prompt.assert_not_called()
        mock_save_event_properties.assert_called()

    @patch('montecarlodata.common.resources.AwsClientWrapper')
    def test_get_dc_resource_props(self, aws_wrapper_mock):
        arn = 'arn:aws:cloudformation:us-east-1:1234:stack/foo/bar'
        collector_props = Box({'stackArn': arn})

        self.maxDiff = None
        self.assertEqual(
            self._service.get_dc_resource_props(collector_props=collector_props),
            DcResourceProperties(
                collector_arn=AWSArn(arn),
                collector_props=collector_props,
                collection_region='us-east-1',
                resources_region='us-east-1',
                collection_client=aws_wrapper_mock(),
                resources_client=aws_wrapper_mock()
            )
        )
        aws_wrapper_mock.assert_has_calls(
            [
                call(profile_name=_SAMPLE_CONFIG.aws_profile, region_name=_SAMPLE_CONFIG.aws_region),
                call(profile_name=_SAMPLE_CONFIG.aws_profile, region_name=_SAMPLE_CONFIG.aws_region)
            ]
        )

    def test_parse_stack_prop_list(self):
        self.assertEqual(
            self._service._parse_stack_prop_list(key='foo', val='bar', struct=[{'foo': 'baz', 'bar': 'qux'}]),
            Box({'baz': 'qux'})
        )


import click
import click_config_file

from montecarlodata import settings
from montecarlodata.common import create_mc_client
from montecarlodata.iac.mc_config_service import MonteCarloConfigService
from montecarlodata.rules.rule_service import RuleService



@click.group(help='Manage monitors.')
def monitors():
    """
    Group for any monitor related subcommands
    """
    pass

@monitors.command(help='Compile monitor configuration.')
@click.option('--project-dir', required=False, help='Base directory of MC project (where montecarlo.yml is located). By default, this is set to the current working directory')
@click.pass_obj
def compile(ctx, project_dir):
    MonteCarloConfigService(config=ctx['config'], project_dir=project_dir).compile()

@monitors.command(help='Compile and apply monitor configuration.')
@click.option('--project-dir', required=False, help='Base directory of MC project (where montecarlo.yml is located). By default, this is set to the current working directory')
@click.option('--namespace', required=True, help='Namespace of monitors configuration.')
@click.option('--dry-run', required=False, default=False, show_default=True, is_flag=True, help='Dry run (just shows planned changes but doesn\'t apply them.')
@click.pass_obj
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def apply(ctx, project_dir, namespace, dry_run):
    MonteCarloConfigService(config=ctx['config'], project_dir=project_dir).apply(namespace, dry_run=dry_run)

@monitors.command(help='Delete monitor configuration.')
@click.option('--project-dir', required=False, help='Base directory of MC project (where montecarlo.yml is located). By default, this is set to the current working directory')
@click.option('--namespace', required=True, help='Namespace of monitors configuration.')
@click.option('--dry-run', required=False, default=False, show_default=True, is_flag=True, help='Dry run (just shows planned changes but doesn\'t apply them.')
@click.pass_obj
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def delete(ctx, project_dir, namespace, dry_run):
    MonteCarloConfigService(config=ctx['config'], project_dir=project_dir).delete(namespace, dry_run=dry_run)

@monitors.command(name='list', help='List monitors ordered by update recency.')
@click.option('--limit', required=False, default=100, type=click.INT, show_default=True, help='Max number of monitors to list.')
@click.option(
    '--rule-type',
    required=True,
    type=click.Choice(RuleService.RULE_TYPES, case_sensitive=False),
    help='The type of rule to list.'
)
@click.pass_obj
def list_monitors(ctx, rule_type, limit):
    if rule_type == RuleService.PSEUDO_RULE_TYPE_CB_COMPATIBLE:
        rule_type = RuleService.RULE_TYPE_CUSTOM_SQL
    RuleService(client=create_mc_client(ctx)).list_rules(rule_type, limit)



import os
import pathlib
from typing import Optional, Dict
from unittest import TestCase
from unittest.mock import Mock

import click
from box import Box

from montecarlodata.common.common import read_as_json_string
from montecarlodata.common.data import MonolithResponse
from montecarlodata.common.user import UserService
from montecarlodata.integrations.onboarding.data_lake.events import EventsOnboardingService
from montecarlodata.integrations.onboarding.fields import EXPECTED_TOGGLE_EVENTS_GQL_RESPONSE_FIELD, \
    GLUE_CONNECTION_TYPE
from montecarlodata.queries.onboarding import TOGGLE_EVENT_MUTATION
from montecarlodata.utils import GqlWrapper, AwsClientWrapper
from tests.test_base_onboarding import _SAMPLE_BASE_OPTIONS
from tests.test_common_user import _SAMPLE_CONFIG, _SAMPLE_DW_ID


class EventsOnboardingTest(TestCase):
    _SAMPLE_CONNECTION_ID = '43'

    def setUp(self) -> None:
        self._user_service_mock = Mock(autospec=UserService)
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)

        self._service = EventsOnboardingService(
            _SAMPLE_CONFIG,
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock
        )

    def test_toggle_events_when_successful(self):
        self.assertTrue(self._test_event_toggle(toggle=True))

    def test_toggle_events_with_mapping_file(self):
        mapping_file = os.path.join(pathlib.Path(__file__).parent.resolve(), 'sample_mapping.json')

        options = {'enable': True, 'mapping_file': mapping_file}
        expected_variables = {'enable': True, 'mapping': read_as_json_string(mapping_file)}

        self.assertTrue(self._test_event_toggle(options=options, expected_variables=expected_variables))

    def test_toggle_events_with_connection_type_upper_case(self):
        options = {'enable': True, 'connection_type': 'GLUE'}
        expected_variables = {'enable': True, 'connection_type': GLUE_CONNECTION_TYPE}

        self.assertTrue(self._test_event_toggle(options=options, expected_variables=expected_variables))

    def test_toggle_events_with_connection_id(self):
        options = {'enable': True, 'connection_id': self._SAMPLE_CONNECTION_ID}

        self.assertTrue(self._test_event_toggle(options=options, expected_variables=options))

    def test_toggle_events_when_unsuccessful(self):
        with self.assertRaises(click.exceptions.Abort):
            self._test_event_toggle(toggle=False, success=False)

    def _test_event_toggle(self, toggle: Optional[bool] = True, success: Optional[bool] = True,
                           options: Optional[Dict] = None, expected_variables: Optional[Dict] = None) -> Optional[bool]:
        # Helper to test event toggle with sample configuration
        if not options:
            options = {**_SAMPLE_BASE_OPTIONS, **{'enable': toggle}}
        if not expected_variables:
            expected_variables = {**options, }

        self._user_service_mock.warehouses = [{'uuid': _SAMPLE_DW_ID}]
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(data=Box({'success': success}))

        status = self._service.toggle_event_configuration(**options)

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=TOGGLE_EVENT_MUTATION,
            operation=EXPECTED_TOGGLE_EVENTS_GQL_RESPONSE_FIELD,
            variables=expected_variables
        )
        return status
